import TDBRecord as tdbra

from subprocess import Popen, PIPE, DEVNULL
from streamlink import Streamlink
from requests import get
from pathlib import Path
import time

session = Streamlink()
urlstore = {
    "twitch": "https://twitch.tv/{user}",
    "afreecatv": "https://play.afreecatv.com/{user}",
}

def worker(user: str, platform: str) -> None:
    """
    Standalone video downloader
    """
    logger = tdbra.data[f"{user}.{platform}"]["logger"]
    logger.setLevel(tdbra.logger.level)
    downloadPath = tdbra.downloadPath / Path(f"{user}.{platform}/")
    downloadPath.mkdir(parents=True, exist_ok=True)
    filename = f"{platform}.{user}.{time.strftime('%y%m%d.%H%M%S')}.mp4"
    m3uPlaylistUri = stream(user, platform)
    logger.info("FFMPEG Downloader started.")

    while True:
        m3ud = get(m3uPlaylistUri)
        if "twitch-ad-quartile" in m3ud.text:
            logger.debug("Waiting for AD to start...")
            time.sleep(2)
            continue
        break

    logger.info("AD finished. Start recording...")
    logger.debug(f"{tdbra.conf['ffmpeg']} -y -hide_banner -loglevel error -i {m3uPlaylistUri} -c copy {downloadPath}/{filename}")
    ffmpeg = Popen([
        tdbra.conf["ffmpeg"],
        "-y",
        "-hide_banner",
        "-loglevel",
        "error",
        "-i",
        m3uPlaylistUri,
        "-c",
        "copy",
        f"{downloadPath}/{filename}"
    ], stdin=PIPE)
    while True:
        if tdbra.data[f"{user}.{platform}"]["exit"]:
            # send "q" to ffmpeg
            ffmpeg.communicate(input=b"q")
            for i in range(10):
                if ffmpeg.poll() is not None:
                    logger.info("FFMPEG Downloader finished with status 0.")
                    del(tdbra.data[f"{user}.{platform}"])
                    return
                time.sleep(1)
            ffmpeg.kill()
            logger.warning("FFMPEG Downloader killed.")
            break
        if ffmpeg.poll() is not None:
            logger.info("FFMPEG Downloader finished with status 0.")
            tdbra.data[f"{user}.{platform}"]["end"] = True
            break
        time.sleep(1)
    del(tdbra.data[f"{user}.{platform}"])
    pass

def stream(user: str, platform: str) -> str:
    if tdbra.conf["remote_streamlink"] and platform == "twitch":
        return get(f"{tdbra.conf['remote_streamlink']}/{user}?quality=best").text
    else:
        return session.streams(urlstore[platform].format(user=user))["best"].url

def checkStatus(user: str, platform: str) -> bool:
    try:
        if session.streams(urlstore[platform].format(user=user)):
            return True
        else:
            return False
    except:
        tdbra.logger.error(f"Failed to check status of {user}.{platform}")
        return False



import TDBRecord as tdbra

import pathlib
import subprocess

DEFAULT_CONFIG = {
    "remote_streamlink": "", # Remote streamlink server, if not set, use local streamlink e.g. "http://www.example.com/api/get"
    "downloadPath": "/mnt/download/",
    "ffmpeg": "ffmpeg", # ffmpeg path, if not set, use PATH ffmpeg
    "users": [
        {
            "name": "ExampleUser1",
            "platform": "twitch | afreecatv"
        },
        {
            "name": "ExampleUser2",
            "platform": "twitch | afreecatv"
        }
    ],
}

PLATFORMS = ["twitch", "afreecatv"]

def check_config():
    if not tdbra.conf["users"]:
        tdbra.logger.error("No user found in config file")
        raise ValueError("No user found in config file")
    
    for user in tdbra.conf["users"]:
        if not user["name"]:
            tdbra.logger.error("User name not found in config file")
            raise ValueError("User name not found in config file")
        if not user["platform"]:
            tdbra.logger.error("User platform not found in config file")
            raise ValueError("User platform not found in config file")
        if user["platform"] not in PLATFORMS:
            tdbra.logger.error("User platform not found in config file")
            raise ValueError("User platform not found in config file")
    
    if not tdbra.conf["downloadPath"]:
        tdbra.logger.error("Download path not found in config file")
        raise ValueError("Download path not found in config file")
    tdbra.downloadPath = pathlib.Path(tdbra.conf["downloadPath"])
    tdbra.downloadPath.mkdir(parents=True, exist_ok=True)

    if not tdbra.conf["ffmpeg"]:
        tdbra.conf["ffmpeg"] = "ffmpeg"
    
    # check ffmpeg is installed
    try:
        subprocess.run([tdbra.conf["ffmpeg"], "-version"], check=True, stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL)
    except subprocess.CalledProcessError:
        tdbra.logger.error("ffmpeg not found")
        raise ValueError("ffmpeg not found")
    except FileNotFoundError:
        tdbra.logger.error("ffmpeg not found")
        raise ValueError("ffmpeg not found")

    pass


import TDBRecord as tdbra

import pathlib
import logging
import click
import json

__version__ = "1.0.5"

@click.version_option(prog_name="TDBRecord", version=__version__)
@click.group()
def main():
    """
    TDB Record Application for OpenSourceUsers
    """
    pass

@main.command()
@click.option("--debug", is_flag=True, help="Enable debug mode")
@click.option("--config", type=click.Path(exists=True, file_okay=True, dir_okay=False, path_type=pathlib.Path), help="Config file", default="config.json")
@click.option("--logfile", type=click.Path(file_okay=True, dir_okay=False, path_type=pathlib.Path), help="Log file", default="record.log")
def start(debug, config, logfile):
    """
    Start TDBRecord server for multiple users.
    """
    if debug: tdbra.logger.setLevel(logging.DEBUG)
    tdbra.conf = json.loads(config.read_text())

    tdbra.logger.addHandler(logging.FileHandler(logfile))
    tdbra.logger.info("TDBRecord started!")

    tdbra.config.check_config()
    tdbra.command.start()

@main.command()
@click.option("--debug", is_flag=True, help="Enable debug mode")
@click.option("--remote-streamlink", is_flag=True, help="Use remote streamlink")
@click.option("--download-path", type=click.Path(exists=True, file_okay=False, dir_okay=True, path_type=pathlib.Path), help="Download file", default="download/")
@click.argument(
    "user",
    type=str,
    required=True,
)
@click.argument(
    "platform",
    type=str,
    required=True,
)
def record(debug, user, platform, remote_streamlink, download):
    """
    Standalone record command. Only record one user.
    """
    if debug: tdbra.loglevel = logging.DEBUG
    tdbra.conf["remote_streamlink"] = remote_streamlink
    tdbra.downloadPath = download.resolve()
    tdbra.command.record(user, platform)

@main.command()
def config():
    """
    Generate config file
    """
    config = pathlib.Path("config.json")
    if config.exists():
        print("Config file already exists")
        return
    config.write_text(json.dumps(tdbra.config.DEFAULT_CONFIG, indent=4))
    tdbra.logger.info("Config file generated")



import TDBRecord as tdbra

from threading import Thread
import time

def start():
    tdbra.logger.debug("TDBRecord Main function started!")

    statusCheckThread = Thread(target=statusChecker)
    statusCheckThread.daemon = True
    statusCheckThread.start()

    tdbra.logger.debug("TDBRecord Status Checker Thread started!")
    while True:
        try:
            command = tdbra.input("> ")
        except KeyboardInterrupt:
            for user in list(tdbra.data.keys()):
                tdbra.data[user]["exit"] = True
            while len(tdbra.data) > 0:
                time.sleep(0.4)
            break
        if command == "":
            continue
        elif command == "exit":
            for user in list(tdbra.data.keys()):
                tdbra.data[user]["exit"] = True
            while len(tdbra.data) > 0:
                time.sleep(0.4)
            break
        elif command == "status":
            users = []
            for user in list(tdbra.data.keys()):
                users.append(user)
            if len(users) == 0:
                tdbra.logger.info("No user is recording")
                continue
            tdbra.logger.info("Users recording:", str(users))
        elif command == "users":
            users = []
            for user in tdbra.conf["users"]:
                users.append(user["name"])
            tdbra.logger.info("Users:", str(users))
        elif command == "help":
            tdbra.logger.info("Available commands: exit, status, users, help")
        else:
            print(exec(command))
    pass

def statusChecker():
    while True:
        for user in tdbra.conf["users"]:
            if f"{user['name']}.{user['platform']}" in tdbra.data: continue

            if tdbra.m3u8.checkStatus(user["name"], user["platform"]):
                if f"{user['name']}.{user['platform']}" not in tdbra.data:
                    recordThread = Thread(target=tdbra.m3u8.worker, args=(user["name"], user["platform"]))
                    recordThread.daemon = True
                    newData = tdbra.create_data(user["name"], user["platform"], recordThread)
                    tdbra.data[f"{user['name']}.{user['platform']}"] = newData
                    recordThread.start()
                    tdbra.logger.debug(f"TDBRecord Record Thread for {user['name']}.{user['platform']} started!")
        
        time.sleep(60)
        pass


import TDBRecord.command
import TDBRecord.m3u8
import TDBRecord.config


from prompt_toolkit import PromptSession, print_formatted_text
from streamlink import Streamlink
from pathlib import Path
import logging

streamlink = Streamlink()
psession = PromptSession()
input = psession.prompt

# Logging
class PromptHandler(logging.StreamHandler):
    def emit(self, record):
        msg = self.format(record)
        print_formatted_text(msg)

logger = logging.getLogger("TDBRecord")
logger.handlers = [PromptHandler()]
logger.handlers[0].setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s # %(message)s'))
lm = "[{user}.{platform}] {msg}"
logger.setLevel(logging.INFO)

def create_logger(name: str, platform: str) -> logging.Logger:
    ulogger = logging.getLogger(name)
    ulogger.handlers = [PromptHandler()]
    ulogger.handlers[0].setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s - [{user}.{platform}] # %(message)s'.format(user=name, platform=platform)))
    ulogger.setLevel(logger.level)
    ulogger.propagate = False
    return ulogger

def create_data(user: str, platform: str, thread = None) -> dict:
    return {
        "exit": False,
        "logger": create_logger(user, platform),
        "thread": thread,
    }




conf = {}
data = {}
downloadPath = Path(".")

""" data
{
    "user.platform": {
        "exit": false,
        "logger": logger,
    }
}
"""



import setuptools
import re

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("TDBRecord/__main__.py", "r") as fh:
    version = re.search(r'__version__ = "(.*)"', fh.read()).group(1)

with open("requirements.txt", "r") as fh:
    install_requires = fh.read().splitlines()

setuptools.setup(
    name="TDBRecord",
    version=version,
    author="onTDB",
    author_email="pypi@ontdb.com",
    description="TDB Record Application",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.5.0",
    install_requires=install_requires,
    entry_points={
        "console_scripts": [
            "tdbrec = TDBRecord.__main__:main",
        ]
    },
)
